import argparse
import os

class ArgParser(object):
    def __init__(self):
        pass

    def parse_args(self):
        parser = argparse.ArgumentParser(
            description="WAN Emulator for Linux")

        parser.add_argument(
            "-n", "--nic",
            choices=os.listdir("/sys/class/net"),
            metavar="interface",
            required=True,
            help="name of the network interface to be impaired"
        )

        direction = parser.add_mutually_exclusive_group(required=False)
        direction.add_argument(
            "--ingress",
            action="store_true",
            help="ingress interface impairment instead of egress"
        )
        direction.add_argument(
            "--bidirectional",
            action="store_true",
            help="bidirectional interface impairment (ingress+egress)"
        )

        parser.add_argument(
            "-i", "--include",
            action="append",
            default=["src=0/0", "src=::/0"],
            help="ip addresses and/or ports to be included in emulation "
                 " (example: --include src=ip,sport=portnum "
                 "--include dst=ip,dport=portnum)"
        )
        parser.add_argument(
            "-e", "--exclude",
            action="append",
            default=["dport=22", "sport=22"],
            help="ip addresses and/or ports to be excluded in emulation "
                 " (example: --exclude src=ip,sport=portnum "
                 "--exclude dst=ip,dport=portnum)"
        )

        parser.add_argument(
            "-d", "--debug",
            action="store_true",
            default=False,
            help="print out debug logging output"
        )

        subparsers = parser.add_subparsers(
            title="impairments",
            dest="subparser_name",
            description="list of available impairments",
            help="valid impairments"
        )

        netem_args = subparsers.add_parser("netem", help="manipulate netem attributes")

        netem_args.add_argument(
            "--loss_ratio",
            type=int,
            default=0,
            help="percentage of packets to be lost"
        )
        netem_args.add_argument(
            "--loss_corr",
            type=int,
            default=0,
            help="correlation factor for the random packet loss"
        )
        netem_args.add_argument(
            "--dupl_ratio",
            type=int,
            default=0,
            help="percentage of packets to be duplicated"
        )
        netem_args.add_argument(
            "--delay",
            type=int,
            default=0,
            help="overall delay for each packet in ms"
        )
        netem_args.add_argument(
            "--jitter",
            type=int,
            default=0,
            help="jitter in ms"
        )
        netem_args.add_argument(
            "--delay_jitter_corr",
            type=int,
            default=0,
            help="correlation factor for the random jitter"
        )
        netem_args.add_argument(
            "--reorder_ratio",
            type=int,
            default=0,
            help="percentage of packets to be reordered"
        )
        netem_args.add_argument(
            "--reorder_corr",
            type=int,
            default=0,
            help="correlation factor for the random reordering"
        )
        netem_args.add_argument(
            "--toggle",
            nargs="+",
            type=int,
            default=[1000000],
            help="toggles impairment on and off on specific intervals (example: "
                 "--toggle 10 3 5 2 will enable impairment for 10 s, turn it off "
                 "for 3 s, turn it on for 5 s, and turn it off for 2 s"
        )

        tbf_args = subparsers.add_parser("tbf", help="manipulate token bucket filter attributes")
        tbf_args.add_argument(
            "--limit",
            type=int,
            default=0,
            help="tbf limit in kb"
        )
        tbf_args.add_argument(
            "--burst",
            type=int,
            default=2000,
            help="amount of tokens in bucket available in bytes"
        )
        tbf_args.add_argument(
            "--latency",
            type=int,
            default=20,
            help="maximum time packets can stay queued before being dropped in ms"
        )
        tbf_args.add_argument(
            "--toggle",
            nargs="+",
            type=int,
            default=[1000000],
            help="toggles impairment on and off on specific intervals (example: "
                 "--toggle 10 3 5 2 will enable impairment for 10 s, turn it off "
                 "for 3 s, turn it on for 5 s, and turn it off for 2 s"
        )

        return parser.parse_args()


import logging
import datetime
import time

from Utils.Subprocess import call, check_call
from Tc.Tbf import TbfCollection, Tbf
from Tc.Filter import Filter, TcFilter
from Tc.Netem import NetemCollection, Duplicate, Delay, Loss, Reorder, Corrupt


class TcWrapper(object):
    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])
        nic = self.nic
        self.nic = 'ifb1' if not self.direction == "Egress" else nic
        self.real_nic = nic

    def tbf(self, operation):
        tbf = Tbf(self.limit,
                  self.burst,
                  self.latency)

        tbf_collection = TbfCollection(self.nic)
        tbf_collection.append(tbf)
        tbf_collection.apply_cmd(operation=operation)

    def netem(self, operation):
        loss = Loss(self.loss_ratio,
                    self.loss_corr)

        dupl = Duplicate(self.dupl_ratio,
                         self.dupl_corr)

        delay = Delay(self.delay,
                      self.jitter,
                      self.delay_jitter_corr)

        reorder = Reorder(self.reorder_ratio,
                          self.reorder_corr)

        corrupt = Corrupt(self.corrupt_ratio)

        netem_collection = NetemCollection(self.nic)
        netem_collection.extend([loss, dupl, delay, reorder, corrupt])
        netem_collection.apply_cmd(operation=operation)

    def initialize(self):
        logging.info("Initializing and cleaning up previous configuration.")

        if not self.direction == "Egress":
            self.ingress_setup()

        call(f"tc qdisc del root dev {self.nic} ")
        check_call(f"tc qdisc add dev {self.nic} root handle 1: prio")

        exclude_filters, exclude_filters_ipv6 = Filter.generate_filters(self.exclude)
        TcFilter(exclude_filters, exclude_filters_ipv6, self.nic, "exclude")

        include_filters, include_filters_ipv6 = Filter.generate_filters(self.include)
        TcFilter(include_filters, include_filters_ipv6, self.nic, "include")

    def teardown(self):
        if not self.direction == "Egress":
            call(f"tc filter del dev {self.real_nic} parent ffff: protocol ip prio 1")
            call(f"tc qdisc del dev {self.real_nic} ingress")
            call("ip link set dev ifb0 down")

        call(f"tc qdisc del root dev {self.nic}")
        logging.info("Network impairment emulation teardown complete.")

    def ingress_setup(self):
        check_call("modprobe ifb")
        check_call(f"ip link set dev {self.nic} up")
        call(f"tc qdisc del dev {self.real_nic} ingress ")
        check_call(f"tc qdisc replace dev {self.real_nic} ingress")
        check_call(f"tc filter replace dev {self.real_nic} parent ffff: protocol "
                   f"ip prio 1 u32 match u32 0 0 flowid 1:1 action mirred egress "
                   f"redirect dev {self.nic}")



import logging

from dataclasses import dataclass
from abc import abstractmethod

from Utils.Subprocess import check_call


class Filter(object):
    _flowid = ""
    @abstractmethod
    def _apply_filter(self, filter_list, ip_version, priority):
        pass

    @staticmethod
    def generate_filters(filter_list):
        filter_strings = []
        filter_strings_ipv6 = []

        for tc_filter in filter_list:
            filter_tokens = tc_filter.split(",")

            try:
                filter_string = ""
                filter_string_ipv6 = ""

                for token in filter_tokens:
                    token_split = token.split("=")
                    key = token_split[0]
                    value = token_split[1]

                    if key == "src" or key == "dst":
                        if "::" in value:
                            filter_string_ipv6 += f"match ip6 {key} {value} "
                        else:
                            filter_string += f"match ip {key} {value} "
                    else:
                        filter_string += f"match ip {key} {value} "
                        filter_string_ipv6 += f"match ip6 {key} {value} "
                    if key == "sport" or key == "dport":
                        filter_string += "0xffff "
                        filter_string_ipv6 += "0xffff "

            except IndexError:
                logging.error("Error: Invalid filter parameters")

            if filter_string:
                filter_strings.append(filter_string)
            if filter_string_ipv6:
                filter_strings_ipv6.append(filter_string_ipv6)

        return filter_strings, filter_strings_ipv6


@dataclass
class TcFilter(Filter):
    _ip: str = None
    _ipv6: str = None
    _nic: str = None
    _type: str = None
    _priority: int = 0

    def __post_init__(self):
        self._type_check(self._type)

        if self._type == "exclude":
            self._flowid = "1:2"
            self._priority = 1
            logging.debug(f"Excluding the following from network impairment:")
        else:
            self._flowid = "1:3"
            self._priority = 3
            logging.debug(f"Including the following for network impairment:")

        self._apply_filter(self._ip, "ip", self._priority)
        self._apply_filter(self._ipv6, "ipv6", self._priority+1)

    def _apply_filter(self, filter_list, ip_version, priority):
        for rule in filter_list:
            filter_cmd = f"tc filter add dev {self._nic} protocol {ip_version} " \
                         f"parent 1:0 prio {priority} u32 {rule}flowid {self._flowid}"
            check_call(filter_cmd)

    def _type_check(self, type):
        _allowed_types = ["include", "exclude"]
        if not (type in _allowed_types):
            raise Exception("Invalid type parameter")
        else:
            return
